/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <jcifs at samba dot org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import jcifs.Config;
import java.io.IOException;
import java.io.InputStream;

class SmbComSessionSetupAndX extends AndXServerMessageBlock {

    static final int BATCH_LIMIT =
            Config.getInt( "jcifs.smb.client.SessionSetupAndX.TreeConnectAndX", 1 );

    byte[] accountPassword, unicodePassword;
    int passwordLength, unicodePasswordLength;
    int negotiatedMaxBufferSize,
        negotiatedMaxMpxCount,
        vcNumber,
        sessionKey;
    String accountName,
        primaryDomain,
        nativeOs,
        nativeLanMan;

    SmbSession session;

    SmbComSessionSetupAndX( SmbSession session, ServerMessageBlock andx ) {
        super( andx );
        command = SMB_COM_SESSION_SETUP_ANDX;
        this.session = session;
    }

    int getBatchLimit( byte command ) {
        return command == SMB_COM_TREE_CONNECT_ANDX ? BATCH_LIMIT : 0;
    }
    int writeParameterWordsWireFormat( byte[] dst, int dstIndex ) {
        int start = dstIndex;

        if( session.transport.server.security == SECURITY_USER &&
                        ( session.auth.hashesExternal ||
                        session.auth.password.length() > 0 )) {
            if( session.transport.server.encryptedPasswords ) {
                // encrypted
                accountPassword = session.auth.getAnsiHash( session.transport.server.encryptionKey );
                unicodePassword = session.auth.getUnicodeHash( session.transport.server.encryptionKey );
                // fix for NTLMv2 or empty NTLM/LM
                passwordLength = accountPassword.length;
                unicodePasswordLength = unicodePassword.length;
            } else if( Config.getBoolean( "jcifs.smb.client.disablePlainTextPasswords", true )) {
                throw new RuntimeException( "Plain text passwords are disabled" );
            } else if( useUnicode ) {
                // plain text
                String password = session.auth.getPassword();
                accountPassword = new byte[0];
                passwordLength = 0;
                unicodePassword = new byte[(password.length() + 1) * 2];
                unicodePasswordLength = writeString( password, unicodePassword, 0 );
            } else {
                // plain text
                String password = session.auth.getPassword();
                accountPassword = new byte[(password.length() + 1) * 2];
                passwordLength = writeString( password, accountPassword, 0 );
                unicodePassword = new byte[0];
                unicodePasswordLength = 0;
            }
        } else {
            // no password in session setup
            passwordLength = unicodePasswordLength = 0;
        }

        negotiatedMaxBufferSize = session.transport.negotiatedMaxBufferSize;
        negotiatedMaxMpxCount = session.transport.negotiatedMaxMpxCount;
        vcNumber = session.transport.client.vcNumber;
        sessionKey = session.transport.client.sessionKey;

        writeInt2( negotiatedMaxBufferSize, dst, dstIndex );
        dstIndex += 2;
        writeInt2( negotiatedMaxMpxCount, dst, dstIndex );
        dstIndex += 2;
        writeInt2( vcNumber, dst, dstIndex );
        dstIndex += 2;
        writeInt4( sessionKey, dst, dstIndex );
        dstIndex += 4;
        writeInt2( passwordLength, dst, dstIndex );
        dstIndex += 2;
        writeInt2( unicodePasswordLength, dst, dstIndex );
        dstIndex += 2;
        dst[dstIndex++] = (byte)0x00;
        dst[dstIndex++] = (byte)0x00;
        dst[dstIndex++] = (byte)0x00;
        dst[dstIndex++] = (byte)0x00;
        writeInt4( session.transport.client.capabilities, dst, dstIndex );
        dstIndex += 4;

        return dstIndex - start;
    }
    int writeBytesWireFormat( byte[] dst, int dstIndex ) {
        int start = dstIndex;

        accountName = session.auth.username.toUpperCase();
        primaryDomain = session.auth.domain.toUpperCase();
        nativeOs = session.transport.client.nativeOs;
        nativeLanMan = session.transport.client.nativeLanMan;

        if( session.transport.server.security == SECURITY_USER &&
                        ( session.auth.hashesExternal ||
                        session.auth.password.length() > 0 )) {
            System.arraycopy( accountPassword, 0, dst, dstIndex, passwordLength );
            dstIndex += passwordLength;
            System.arraycopy( unicodePassword, 0, dst, dstIndex, unicodePasswordLength );
            dstIndex += unicodePasswordLength;
        }
        if( useUnicode ) {
            // at least NT 4 observed needing this only with unicode
            dst[dstIndex++] = (byte)'\0';
        }

        dstIndex += writeString( accountName, dst, dstIndex );
        dstIndex += writeString( primaryDomain, dst, dstIndex );
        dstIndex += writeString( nativeOs, dst, dstIndex );
/*
        // at least NT 4 observed with 2 zero bytes here
        dst[dstIndex++] = (byte)0x00;
        dst[dstIndex++] = (byte)0x00;
This still isn't quite right.
*/
        dstIndex += writeString( nativeLanMan, dst, dstIndex );

        return dstIndex - start;
    }
    int readParameterWordsWireFormat( byte[] buffer, int bufferIndex ) {
        return 0;
    }
    int readBytesWireFormat( byte[] buffer, int bufferIndex ) {
        return 0;
    }
    int readBytesDirectWireFormat( InputStream in, int byteCount ) throws IOException {
        return 0;
    }
    public String toString() {
        String result = new String( "SmbComSessionSetupAndX[" +
            super.toString() +
            ",maxBufferSize=" + negotiatedMaxBufferSize +
            ",maxMpxCount=" + negotiatedMaxMpxCount +
            ",vcNumber=" + vcNumber +
            ",sessionKey=" + sessionKey +
            ",passwordLength=" + passwordLength +
            ",unicodePasswordLength=" + unicodePasswordLength +
            ",capabilities=" + session.transport.client.capabilities +
            ",accountName=" + accountName +
            ",primaryDomain=" + primaryDomain +
            ",nativeOs=" + nativeOs +
            ",nativeLanMan=" + nativeLanMan + "]" );
        return result;
    }
}
