/* jcifs smb client library in Java
 * Copyright (C) 2002  "Michael B. Allen" <jcifs at samba dot org>
 *                   "Eric Glass" <jcifs at samba dot org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.ntlmssp;

import java.io.ByteArrayOutputStream;
import java.io.IOException;

import java.net.UnknownHostException;

import java.util.Arrays;

import jcifs.Config;

import jcifs.netbios.NbtAddress;

/**
 * Represents an NTLMSSP Type-2 message.
 */
public class Type2Message extends NtlmMessage {

    private byte[] challenge;

    private String target;

    private byte[] context;

    private byte[] targetInformation;

    /**
     * Creates a Type-2 message using default values from the current
     * environment.
     */
    public Type2Message() {
        this(getDefaultFlags(), null, null);
    }

    /**
     * Creates a Type-2 message in response to the given Type-1 message
     * using default values from the current environment.
     *
     * @param type1 The Type-1 message which this represents a response to.
     */
    public Type2Message(Type1Message type1) {
        this(type1, null, null);
    }

    /**
     * Creates a Type-2 message in response to the given Type-1 message.
     *
     * @param type1 The Type-1 message which this represents a response to.
     * @param challenge The challenge from the domain controller/server.
     * @param target The authentication target.
     */
    public Type2Message(Type1Message type1, byte[] challenge, String target) {
        this(getDefaultFlags(type1), challenge, (type1 != null &&
                target == null && type1.getFlag(NTLMSSP_REQUEST_TARGET)) ?
                        getDefaultDomain() : target);
    }

    /**
     * Creates a Type-2 message with the specified parameters.
     *
     * @param flags The flags to apply to this message.
     * @param challenge The challenge from the domain controller/server.
     * @param target The authentication target.
     */
    public Type2Message(int flags, byte[] challenge, String target) {
        setFlags(flags);
        setChallenge(challenge);
        setTarget(target);
        if (target != null) setTargetInformation(getDefaultTargetInformation());
    }

    /**
     * Creates a Type-2 message using the given raw Type-2 material.
     *
     * @param material The raw Type-2 material used to construct this message.
     * @throws IOException If an error occurs while parsing the material.
     */
    public Type2Message(byte[] material) throws IOException {
        parse(material);
    }

    /**
     * Returns the challenge for this message.
     *
     * @return A <code>byte[]</code> containing the challenge.
     */
    public byte[] getChallenge() {
        return challenge;
    }

    /**
     * Sets the challenge for this message.
     *
     * @param challenge The challenge from the domain controller/server.
     */
    public void setChallenge(byte[] challenge) {
        this.challenge = challenge;
    }

    /**
     * Returns the authentication target.
     *
     * @return A <code>String</code> containing the authentication target.
     */
    public String getTarget() {
        return target;
    }

    /**
     * Sets the authentication target.
     *
     * @param target The authentication target.
     */
    public void setTarget(String target) {
        this.target = target;
    }

    /**
     * Returns the target information block.
     *
     * @return A <code>byte[]</code> containing the target information block.
     * The target information block is used by the client to create an
     * NTLMv2 response.
     */ 
    public byte[] getTargetInformation() {
        return targetInformation;
    }

    /**
     * Sets the target information block.
     * The target information block is used by the client to create
     * an NTLMv2 response.
     * 
     * @param targetInformation The target information block.
     */
    public void setTargetInformation(byte[] targetInformation) {
        this.targetInformation = targetInformation;
    }

    /**
     * Returns the local security context.
     *
     * @return A <code>byte[]</code> containing the local security
     * context.  This is used by the client to negotiate local
     * authentication.
     */
    public byte[] getContext() {
        return context;
    }

    /**
     * Sets the local security context.  This is used by the client
     * to negotiate local authentication.
     *
     * @param context The local security context.
     */
    public void setContext(byte[] context) {
        this.context = context;
    }

    public byte[] toByteArray() {
        try {
            ByteArrayOutputStream type2 = new ByteArrayOutputStream();
            type2.write("NTLMSSP\u0000".getBytes("US-ASCII"));
            writeULong(type2, 2);
            int flags = getFlags();
            String targetString = getTarget();
            byte[] target = new byte[0];
            if ((flags & (NTLMSSP_TARGET_TYPE_DOMAIN |
                    NTLMSSP_TARGET_TYPE_SERVER |
                            NTLMSSP_TARGET_TYPE_SHARE)) != 0) {
                if (targetString != null && targetString.length() != 0) {
                    target =
                            (flags & NTLMSSP_NEGOTIATE_UNICODE) != 0 ?
                                targetString.getBytes("UnicodeLittleUnmarked") :
                                targetString.toUpperCase().getBytes(
                                        getOEMEncoding());
                } else {
                    flags &= (0xffffffff ^ (NTLMSSP_TARGET_TYPE_DOMAIN |
                            NTLMSSP_TARGET_TYPE_SERVER |
                                    NTLMSSP_TARGET_TYPE_SHARE));
                }
            }
            int offset = 32; // start of data block
            byte[] context = getContext();
            byte[] targetInformation = getTargetInformation();
            if (targetInformation != null) {
                flags ^= NTLMSSP_NEGOTIATE_TARGET_INFO;
                // empty context is needed for padding when t.i. is supplied.
                if (context == null) context = new byte[8];
            }
            if (context != null) {
                offset += 8;
                if (targetInformation != null) offset += 8;
            }
            int length = target.length;
            writeUShort(type2, length);
            writeUShort(type2, length);
            writeULong(type2, offset);
            writeULong(type2, flags);
            byte[] bytes = new byte[8];
            byte[] challenge = getChallenge();
            if (challenge != null) {
                System.arraycopy(challenge, 0, bytes, 0,
                        Math.min(challenge.length, 8));
            }
            type2.write(bytes);
            if (context != null) {
                bytes = new byte[8];
                System.arraycopy(context, 0, bytes, 0,
                        Math.min(context.length, 8));
                type2.write(bytes);
                if (targetInformation != null) {
                    writeUShort(type2, targetInformation.length);
                    writeUShort(type2, targetInformation.length);
                    writeULong(type2, offset + length);
                }
            }
            type2.write(target);
            if (targetInformation != null) type2.write(targetInformation);
            return type2.toByteArray();
        } catch (IOException ex) {
            throw new IllegalStateException(ex.getMessage());
        }
    }

    public String toString() {
        String target = getTarget();
        byte[] challenge = getChallenge();
        byte[] context = getContext();
        byte[] targetInformation = getTargetInformation();
        int flags = getFlags();
        StringBuffer buffer = new StringBuffer();
        if (target != null) {
            buffer.append("target: ").append(target);
        }
        if (challenge != null) {
            if (buffer.length() > 0) buffer.append("; ");
            buffer.append("challenge: ");
            buffer.append("0x");
            for (int i = 0; i < challenge.length; i++) {
                buffer.append(Integer.toHexString((challenge[i] >> 4) & 0x0f));
                buffer.append(Integer.toHexString(challenge[i] & 0x0f));
            }
        }
        if (context != null) {
            if (buffer.length() > 0) buffer.append("; ");
            buffer.append("context: ");
            buffer.append("0x");
            for (int i = 0; i < context.length; i++) {
                buffer.append(Integer.toHexString((context[i] >> 4) & 0x0f));
                buffer.append(Integer.toHexString(context[i] & 0x0f));
            }
        }
        if (targetInformation != null) {
            if (buffer.length() > 0) buffer.append("; ");
            buffer.append("targetInformation: ");
            buffer.append("0x");
            for (int i = 0; i < targetInformation.length; i++) {
                buffer.append(Integer.toHexString((targetInformation[i] >> 4) &
                        0x0f));
                buffer.append(Integer.toHexString(targetInformation[i] & 0x0f));
            }
        }
        if (flags != 0) {
            if (buffer.length() > 0) buffer.append("; ");
            buffer.append("flags: ");
            buffer.append("0x");
            buffer.append(Integer.toHexString((flags >> 28) & 0x0f));
            buffer.append(Integer.toHexString((flags >> 24) & 0x0f));
            buffer.append(Integer.toHexString((flags >> 20) & 0x0f));
            buffer.append(Integer.toHexString((flags >> 16) & 0x0f));
            buffer.append(Integer.toHexString((flags >> 12) & 0x0f));
            buffer.append(Integer.toHexString((flags >> 8) & 0x0f));
            buffer.append(Integer.toHexString((flags >> 4) & 0x0f));
            buffer.append(Integer.toHexString(flags & 0x0f));
        }
        return buffer.toString();
    }

    /**
     * Returns the default flags for a generic Type-2 message in the
     * current environment.
     *
     * @return An <code>int</code> containing the default flags.
     */
    public static int getDefaultFlags() {
        int flags = NTLMSSP_NEGOTIATE_NTLM;
        flags |= Config.getBoolean("jcifs.smb.client.useUnicode", true) ?
                NTLMSSP_NEGOTIATE_UNICODE : NTLMSSP_NEGOTIATE_OEM;
        return flags;
    }

    /**
     * Returns the default flags for a Type-2 message created in response
     * to the given Type-1 message in the current environment.
     *
     * @return An <code>int</code> containing the default flags.
     */
    public static int getDefaultFlags(Type1Message type1) {
        if (type1 == null) return getDefaultFlags();
        int flags = NTLMSSP_NEGOTIATE_NTLM;
        int type1Flags = type1.getFlags();
        flags |= ((type1Flags & NTLMSSP_NEGOTIATE_UNICODE) != 0) ?
                NTLMSSP_NEGOTIATE_UNICODE : NTLMSSP_NEGOTIATE_OEM;
        if ((type1Flags & NTLMSSP_REQUEST_TARGET) != 0) {
            String domain = getDefaultDomain();
            if (domain != null) {
                flags |= NTLMSSP_REQUEST_TARGET | NTLMSSP_TARGET_TYPE_DOMAIN;
            }
        }
        return flags;
    }

    /**
     * Returns the default domain from the current environment.
     *
     * @return A <code>String</code> containing the domain.
     */
    public static String getDefaultDomain() {
        return Config.getProperty("jcifs.smb.client.domain", null);
    }

    public static byte[] getDefaultTargetInformation() {
        ByteArrayOutputStream targetInformation = new ByteArrayOutputStream();
        String domain = getDefaultDomain();
        if (domain != null) {
            try {
                byte[] bytes = domain.getBytes("UnicodeLittleUnmarked");
                writeUShort(targetInformation, 2);
                writeUShort(targetInformation, bytes.length);
                targetInformation.write(bytes);
            } catch (IOException ex) { }
        }
        String server = null;
        try {
            server = NbtAddress.getLocalHost().getHostName();
        } catch (UnknownHostException ex) { }
        if (server != null) {
            try {
                byte[] bytes = server.getBytes("UnicodeLittleUnmarked");
                writeUShort(targetInformation, 1);
                writeUShort(targetInformation, bytes.length);
                targetInformation.write(bytes);
            } catch (IOException ex) { }
        }
        try {
            targetInformation.write(new byte[4]); // terminator subblock
        } catch (IOException ex) { }
        return targetInformation.toByteArray();
    }

    private void parse(byte[] material) throws IOException {
        int flags = readULong(material, 20);
        setFlags(flags);
        String target = null;
        byte[] bytes = readSecurityBuffer(material, 12);
        if (bytes.length != 0) {
            target = new String(bytes,
                    ((flags & NTLMSSP_NEGOTIATE_UNICODE) != 0) ?
                            "UnicodeLittleUnmarked" : getOEMEncoding());
        }
        setTarget(target);
        bytes = new byte[8];
        System.arraycopy(material, 24, bytes, 0, 8);
        if (Arrays.equals(bytes, new byte[8])) bytes = null;
        setChallenge(bytes);
        int offset = readULong(material, 16); // offset of targetname start
        if (offset == 32 || material.length == 32) return;
        bytes = new byte[8];
        System.arraycopy(material, 32, bytes, 0, 8);
        if (Arrays.equals(bytes, new byte[8])) bytes = null;
        setContext(bytes);
        if (offset == 40 || material.length == 40) return;
        bytes = readSecurityBuffer(material, 40);
        if (bytes.length != 0) setTargetInformation(bytes);
    }

}
